;; Generated package description from llm.el  -*- no-byte-compile: t -*-
(define-package "llm" "0.9.1" "Interface to pluggable llm backends" '((emacs "28.1")) :commit "bed2cb774d2836a8992ad7cbf821e613d7d14ce1" :authors '(("Andrew Hyatt" . "ahyatt@gmail.com")) :maintainer '("Andrew Hyatt" . "ahyatt@gmail.com") :url "https://github.com/ahyatt/llm")
